from os import PathLike
from typing import Union, Any
from typing_extensions import Self
from multiprocessing import cpu_count as _cpu_count


def Bool(s: str) -> bool:
    return s.lower() == "true"


_required_configs: dict = {
    "data": ("./data", str),
    "batch_size": ("16", int),
    "num_works": (str(_cpu_count()), int),
    "gpu_acceleration": ("False", Bool),
    "device": ("0", int),
}


class Config(object):
    def __init__(self, filename: Union[str, PathLike]):
        self._filename: Union[str, PathLike] = filename
        self._config: dict = {}

    def __contains__(self, key: str) -> bool:
        return key in self._config.keys()

    def set(self, key: str, value: str):
        if hasattr(self, key):
            raise KeyError(f"{key} is used.")
        setattr(self, key, value)
        self._config[key] = value

    def load(self) -> Self:
        with open(self._filename, "r") as f:
            lines = f.readlines()
            last = ""
            for line in lines:
                line = (last + line).replace(" ", "").replace("\n", "")
                if last != "" and last[-1] != ":" and line[0] != ":":
                    last += line
                    continue
                if line[-1] == ":":
                    last += line
                    continue
                sp = line.split(":")
                if len(sp) == 1:
                    last = line
                    continue
                key, val = sp[0], "".join(sp[1:])
                self.set(key, val)
                last = ""
            if last != "":
                raise SyntaxError("Config file didn't end.")
        return self

    def get(self, key: str, must_exists: bool = False, required_type: type = str, default_val: Any = None) \
            -> Union[Any, None]:
        if key not in self._config.keys():
            if must_exists:
                raise KeyError(f"No such configuration: \"{key}\".")
            return default_val
        val = self._config[key]
        return val if required_type == str else required_type(val)

    def get_predefined(self, key: str, must_exists: bool = False) -> Union[Any, None]:
        required_type = str
        if key in _required_configs.keys():
            required_type = _required_configs[key][1]
        return self.get(key, must_exists, required_type)


def new_config(filename: Union[str, PathLike]) -> Config:
    config = Config(filename).load()
    # Check required configurations
    for req_cfg_key in _required_configs.keys():
        if req_cfg_key not in config:
            config.set(req_cfg_key, _required_configs[req_cfg_key][0])
    return config



from os import PathLike
from typing import Union
from copy import copy as _copy
from typing_extensions import Self
from torch.nn import Module as _Module
from torch.nn import modules as _modules
from torch.optim import Optimizer as _Optimizer
from torch import Tensor as _Tensor, save as _save, load as _load

from papercandy.universal import network as _network


class DataCompound(_network.DataCompound):
    def __init__(self, data: _Tensor, target: _Tensor):
        super(DataCompound, self).__init__(data, target, d_type=_Tensor)

    def unpack(self) -> [_Tensor, _Tensor]:
        return super(DataCompound, self).unpack()

    def gpu(self) -> Self:
        o = _copy(self)
        o.data, o.target = self.data.gpu(), self.target.gpu()
        return o


LayerInfo = _network.LayerInfo
LayerInfoList = _network.LayerInfoList


class NetworkC(_network.NetworkC):
    def __init__(self, network: _Module):
        self._network: _Module = network

    def gpu(self) -> Self:
        o = _copy(self)
        o._network = self._network.cuda()
        return o

    def get(self) -> _Module:
        return self._network

    def save(self, filename: Union[str, PathLike]):
        _save(self.get(), filename)

    def load(self, filename: Union[str, PathLike]) -> Self:
        self._network.load_state_dict(_load(filename))
        return self

    def structure(self) -> LayerInfoList[LayerInfo]:
        lil = LayerInfoList()
        for val in self._network.__dict__["_modules"].values():
            if isinstance(val, _modules.Sequential):
                lil += self.reflect_sequential(val)
            layer_info = self.layer2layer_info(val)
            if layer_info is not None:
                lil.append(layer_info)
        return lil

    def reflect_sequential(self, seq: _modules.Sequential) -> LayerInfoList[LayerInfo]:
        lil = LayerInfoList()
        for layer in seq:
            if isinstance(layer, _modules.Sequential):
                lil += self.reflect_sequential(layer)
            layer_info = self.layer2layer_info(layer)
            if layer_info is not None:
                lil.append(layer_info)
        return lil

    @staticmethod
    def layer2layer_info(layer: _Module) -> Union[LayerInfo, None]:
        angle = 23
        if isinstance(layer, _modules.conv._ConvNd):
            return LayerInfo(1200, 1200, angle, "Conv", f"{layer.in_channels}(in)x{layer.out_channels}(out)")
        if isinstance(layer, _modules.pooling._MaxPoolNd):
            return LayerInfo(800, 800, angle, "Pooling")
        if isinstance(layer, _modules.dropout._DropoutNd):
            return LayerInfo(800, 800, angle, "Dropout")
        if isinstance(layer, _modules.batchnorm._BatchNorm):
            return LayerInfo(800, 800, angle, "BatchNorm")
        if isinstance(layer, _modules.ReLU):
            return LayerInfo(400, 400, angle, "Relu")
        return None


class LossFunctionC(_network.LossFunctionC):
    def __init__(self, loss_function: _Module):
        self._loss_function: _Module = loss_function

    def gpu(self) -> Self:
        o = _copy(self)
        o._loss_function = self._loss_function.cuda()
        return o

    def get(self) -> _Module:
        return self._loss_function

    def save(self, filename: Union[str, PathLike]):
        pass

    def load(self, filename: Union[str, PathLike]) -> Self:
        return self


class OptimizerC(_network.OptimizerC):
    def __init__(self, optimizer: _Optimizer):
        self._optimizer: _Optimizer = optimizer

    def gpu(self) -> Self:
        return self

    def get(self) -> _Optimizer:
        return self._optimizer

    def save(self, filename: Union[str, PathLike]):
        _save(self.get(), filename)

    def load(self, filename: Union[str, PathLike]) -> Self:
        self._optimizer.load_state_dict(_load(filename))
        return self
